/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef PNclMDUtils_hpp
#define PNclMDUtils_hpp

#include <vector>
// FIXME: ParCALFuncs.h is not needed here - move reqd Ncl hdrs here
extern "C"{
#include "ParCALFuncs.h"
}
#include "ParCALCore.hpp"
#include "ParCALVarSHnd.hpp"

namespace parncl{

void PNclPrintSel(const NclSelectionRecord *sel);
/* Update new_sel to be selection of orig_sel */
NclSelectionRecord *PNclSelOfSel(const NclSelectionRecord *orig_sel, NclSelectionRecord *new_sel);

/* Deep copy a selector */
NclSelectionRecord *PNclSelCopy(NclSelectionRecord *to_sel, const NclSelectionRecord *from_sel);
NclSelectionRecord *PNclSelCopy(NclSelectionRecord *to_sel, const NclSelectionRecord *from_sel, const size_t from_sel_index);

/* Get selector size */
void PNclGetSelSz(NclSelectionRecord *sel, std::vector<size_t> &sel_sz);

static inline void PNclSFCopyInitPlaceVal(NclSelectionRecord *sel, std::vector<size_t> &place_val, std::vector<size_t> &place_val_vec_idx)
{
	for(int i=0; i<sel->n_entries; i++){
		if((sel->selection[i].sel_type == Ncl_SUB_ALL) || (sel->selection[i].sel_type == Ncl_SUBSCR)){
			place_val[i] = sel->selection[i].u.sub.start;
		}
		else{
			// Ncl_VECSUBSCR
			place_val_vec_idx[i] = 0;
			place_val[i] = sel->selection[i].u.vec.ind[place_val_vec_idx[i]];
		}
	}
}

static inline void PNclSFCopyIncPlaceVal(NclSelectionRecord *sel, std::vector<size_t> &place_val, std::vector<size_t> &place_val_vec_idx, size_t n_dim_blocks)
{
	size_t n_dims = sel->n_entries;

	for(int j=(n_dims-n_dim_blocks)-1; j>=0; j--){
		if((sel->selection[j].sel_type == Ncl_SUB_ALL) || (sel->selection[j].sel_type == Ncl_SUBSCR)){
			size_t stride = sel->selection[j].u.sub.stride;
			size_t max = sel->selection[j].u.sub.finish;
			if((place_val[j] + stride)<= max){
				place_val[j] += stride;	
				break;
			}
			else{
				size_t min = sel->selection[j].u.sub.start;
				place_val[j] = min;
			}
		}
		else{
			// Ncl_VECSUBSCR
			place_val_vec_idx[j] += 1;
			if(place_val_vec_idx[j] < sel->selection[j].u.vec.n_ind){
				place_val[j] = sel->selection[j].u.vec.ind[place_val_vec_idx[j]];
				break;
			}
			else{
				place_val[j] = sel->selection[j].u.vec.ind[0];
				place_val_vec_idx[j] = 0;
			}
		}
	}
}

template <typename T>
void *PNclDefaultSFCopy(T *to_buf, NclSelectionRecord *to_sel, std::vector<size_t> &to_weights, T *from_buf, NclSelectionRecord *from_sel, std::vector<size_t> &from_weights, const size_t n_elems)
{
	const int n_dims = to_sel->n_entries;	
	std::vector<size_t> to_place_val(n_dims, 0);
	std::vector<size_t> from_place_val(n_dims, 0);
	std::vector<size_t> to_place_val_vec_idx(n_dims, 0);
	std::vector<size_t> from_place_val_vec_idx(n_dims, 0);

	/*
	printf(" DSF(%d %d %d %d) (%d %d %d %d)\n", to_weights[0], to_weights[1], to_weights[2], to_weights[3], from_weights[0], from_weights[1], from_weights[2], from_weights[3]); fflush(stdout);
	*/

	PNclSFCopyInitPlaceVal(to_sel, to_place_val, to_place_val_vec_idx);
	PNclSFCopyInitPlaceVal(from_sel, from_place_val, from_place_val_vec_idx);

	/* Copy from q to p */
	T *p = to_buf, *q = from_buf;

	for(size_t i=0; i<n_elems; i++){
		size_t to_idx = 0, from_idx = 0;
		for(int k=0; k<n_dims; k++){
			to_idx += to_weights[k] * to_place_val[k];
			from_idx += from_weights[k] * from_place_val[k];
		}
		/*
		printf(" (%d %d %d %d) = %d\n", to_place_val[0], to_place_val[1], to_place_val[2], (to_place_val.size() >= 4) ? to_place_val[3] : -1, to_idx); fflush(stdout); 
		printf(" (%d %d %d %d) = %d (%f)\n", from_place_val[0], from_place_val[1], from_place_val[2], (from_place_val.size() >= 4) ? from_place_val[3] : -1, from_idx, *(q+from_idx)); fflush(stdout); 
		*/
		if((p+to_idx) != (q+from_idx)){
			*(p+to_idx) = *(q+from_idx);
		}

		PNclSFCopyIncPlaceVal(to_sel, to_place_val, to_place_val_vec_idx, 0);
		PNclSFCopyIncPlaceVal(from_sel, from_place_val, from_place_val_vec_idx, 0);
	}

	return (void *)to_buf;
}

template <typename T>
void *PNclBlockSFCopy(T *to_buf, NclSelectionRecord *to_sel, std::vector<size_t> &to_weights, T *from_buf, NclSelectionRecord *from_sel, std::vector<size_t> &from_weights, const int n_blocks, const int n_dim_blocks)
{
	const int n_dims = to_sel->n_entries;	
	std::vector<size_t> from_place_val(n_dims, 0);
	std::vector<size_t> to_place_val(n_dims, 0);
	std::vector<size_t> from_place_val_vec_idx(n_dims, 0);
	std::vector<size_t> to_place_val_vec_idx(n_dims, 0);
	size_t block_sz = 1;

	/*
	printf(" BSF(%d %d %d %d) : (%d %d %d %d) \n", to_weights[0], to_weights[1], to_weights[2], to_weights[3], from_weights[0], from_weights[1], from_weights[2], from_weights[3]); fflush(stdout);
	*/

	PNclSFCopyInitPlaceVal(to_sel, to_place_val, to_place_val_vec_idx);
	PNclSFCopyInitPlaceVal(from_sel, from_place_val, from_place_val_vec_idx);

	/* Contig blocks are right-most dims - and SUB_ALL selectors */
	for(int i=0; i<n_dim_blocks; i++){
		int dim_in_block = (n_dims - 1) - i;
		block_sz *= (to_sel->selection[dim_in_block].u.sub.finish - to_sel->selection[dim_in_block].u.sub.start + 1);
	}

	/* Copy from q to p */
	T *p = to_buf, *q = from_buf;

	for(int i=0; i<n_blocks; i++){
		size_t to_idx = 0, from_idx = 0;
		for(int k=0; k<(n_dims - n_dim_blocks); k++){
			to_idx += to_weights[k] * to_place_val[k];
			from_idx += from_weights[k] * from_place_val[k];
		}

		/*
		printf(" (%d %d %d %d) = %d\n", to_place_val[0], to_place_val[1], to_place_val[2], (to_place_val.size() >= 4) ? to_place_val[3] : -1, to_idx); fflush(stdout); 
		printf(" (%d %d %d %d) = %d\n", from_place_val[0], from_place_val[1], from_place_val[2], (from_place_val.size() >= 4) ? from_place_val[3] : -1, from_idx); fflush(stdout); 
		*/
		if((p+to_idx) != (q+from_idx)){
			memmove((void *)(p+to_idx), (void *)(q+from_idx), block_sz * sizeof(T));
		}
		PNclSFCopyIncPlaceVal(to_sel, to_place_val, to_place_val_vec_idx, n_dim_blocks);
		PNclSFCopyIncPlaceVal(from_sel, from_place_val, from_place_val_vec_idx, n_dim_blocks);
	}

	return (void *)to_buf;
}

template<typename T>
void *PNclSlideAndFill(const std::vector<size_t> &to_var_dims, NclSelectionRecord *to_sel,T *to_buf, size_t to_buf_sz, const std::vector<size_t> &from_var_dims, NclSelectionRecord *from_sel, T *from_buf, size_t from_buf_sz)
{
	int n_blocks = 0, n_dim_blocks = 0;

	if((from_sel == NULL) || (from_sel->n_entries == 0)) return to_buf;

	const int n_dims = from_sel->n_entries;	
	assert(to_sel->n_entries == n_dims);

	std::vector<size_t> from_weights(n_dims, 1);
	std::vector<size_t> to_weights(n_dims, 1);

	// FIXME: We need to convert negative strides to positive ones before continuing
	// FIXME: Avoid modifying sel - this is the easiest way as of now
	for(int i=0; i<n_dims; i++){
		if(from_sel->selection[i].sel_type == Ncl_SUB_ALL){
			from_sel->selection[i].u.sub.start = 0;
			from_sel->selection[i].u.sub.finish = from_var_dims[i] - 1;
			from_sel->selection[i].u.sub.stride = 1;
		}
		if(to_sel->selection[i].sel_type == Ncl_SUB_ALL){
			to_sel->selection[i].u.sub.start = 0;
			to_sel->selection[i].u.sub.finish = to_var_dims[i] - 1;
			to_sel->selection[i].u.sub.stride = 1;
		}
	}

	/* Find num of elements in the final buf */
	size_t n_elems = 0;
	n_blocks = 1;
	std::vector<size_t> from_n_elems_dim(n_dims, 0);
	std::vector<size_t> to_n_elems_dim(n_dims, 0);
	PNclGetSelSz(from_sel, from_n_elems_dim);
	PNclGetSelSz(to_sel, to_n_elems_dim);

	for(int i=0; i<n_dims; i++){
		size_t n_elems_dim = to_n_elems_dim[i];

		if(n_elems_dim != 0){
			if(n_elems != 0){
				n_elems *= n_elems_dim;
			}
			else{
				n_elems = n_elems_dim;
			}
		}
		else{
			/* Nothing to be done */
			return ((void *)to_buf);
		}
		if((from_sel->selection[i].sel_type == Ncl_SUB_ALL) &&
			(to_sel->selection[i].sel_type == Ncl_SUB_ALL) &&
			(from_n_elems_dim[i] == to_n_elems_dim[i])){
			n_dim_blocks++;
		}
		else{
			n_dim_blocks = 0;
			n_blocks = n_elems;
		}
	}

	//printf("n_elems = %d, n_blocks = %d, n_dim_blocks = %d\n", n_elems, n_blocks, n_dim_blocks); fflush(stdout);

	/* Initialize weights */
	for(int i=n_dims-2; i>=0; i--){
		from_weights[i] = from_weights[i+1] * from_var_dims[i+1];
		to_weights[i] = to_weights[i+1] * to_var_dims[i+1];
	}

	if(n_dim_blocks != 0){
		return PNclBlockSFCopy(to_buf, to_sel, to_weights, from_buf, from_sel, from_weights, n_blocks, n_dim_blocks);
	}
	else{
		return PNclDefaultSFCopy(to_buf, to_sel, to_weights, from_buf, from_sel, from_weights, n_elems);
	}
}

template<typename T>
void *PNclSlideAndFill(const std::vector<size_t> &from_var_dim_sz, NclSelectionRecord *from_sel, T *buf, size_t buf_sz)
{
	if((from_sel == NULL) || (from_sel->n_entries == 0)){
		return buf;
	}

	NclSelectionRecord *to_sel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
	if(!to_sel){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to alloc mem for to var selector");
		return NULL;
	}
	to_sel->n_entries = from_var_dim_sz.size();
	std::vector<size_t> to_var_dim_sz(from_var_dim_sz.size(), 0);
	PNclGetSelSz(from_sel, to_var_dim_sz);

	for(size_t i=0; i<to_sel->n_entries; i++){
		to_sel->selection[i].sel_type = Ncl_SUB_ALL;
		to_sel->selection[i].u.sub.start = 0;
		to_sel->selection[i].u.sub.finish = to_var_dim_sz[i] - 1;
	}
	return PNclSlideAndFill(to_var_dim_sz, to_sel, buf, buf_sz, from_var_dim_sz, from_sel, buf, buf_sz);
}

} // namespace parncl

#endif // PNclMDUtils_hpp
