/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <string>
#include <vector>
#include <map>

#define MPICH_SKIP_MPICXX
#define OMPI_SKIP_MPICXX
extern "C" {
#include "ParCALFuncs.h"
}
#include "mpi.h"
#ifndef MPI_INCLUDED
#define MPI_INCLUDED
#endif
/*
#include "parcal/fileinfo.hpp"
#include "parcal/pcdim.hpp"
#include "parcal/algorithm.hpp"
#include "parcal/exception.hpp"
*/
#include "moab/Core.hpp"
#include "pargal/pargal.hpp"
#include "ParCALCore.hpp"

#include "ParCALVarSHnd.hpp"
#include "MBMesh.hpp"
#include "ParCALMeshVal.hpp"
#include "PNclMBUtils.hpp"
#include "ParCALIAlgo.hpp"

/* There is a separate ParCAL handle per file */
typedef struct _ParCALHnd{
	parncl::MBMesh *mesh;
	int fnames_idx;
} ParCALHnd;

MPI_Comm pcal_gcomm = MPI_COMM_NULL;
int pcal_grank = -1;
moab::Interface *mb_gcore = NULL;
moab::ParallelComm *mb_gcore_pcomm = NULL;
NclQuark time_gq, lev_gq;

std::vector<NclQuark> parcal_supported_funcs;

std::map<std::string, std::pair<nc_type, NclBasicDataTypes> > parcal_to_nc_gtype_map;
std::map<std::string, moab::EntityType> parcal_loc_to_mb_gtype_map;

static void ParCALToNCNclTypeMapInit(void )
{
	parcal_to_nc_gtype_map[std::string("byte")] =  std::make_pair(NC_BYTE, NCL_byte);
	parcal_to_nc_gtype_map[std::string("char")] =  std::make_pair(NC_CHAR, NCL_string);
	parcal_to_nc_gtype_map[std::string("short")] =  std::make_pair(NC_SHORT, NCL_short);
	parcal_to_nc_gtype_map[std::string("int")] =  std::make_pair(NC_INT, NCL_int);
	parcal_to_nc_gtype_map[std::string("float")] =  std::make_pair(NC_FLOAT, NCL_float);
	parcal_to_nc_gtype_map[std::string("double")] =  std::make_pair(NC_DOUBLE, NCL_double);
}

static void ParCALLocToMBTypeMapInit(void )
{
	parcal_loc_to_mb_gtype_map[std::string("QUAD")] = moab::MBQUAD;
	parcal_loc_to_mb_gtype_map[std::string("VERTEX")] = moab::MBVERTEX;
}

/* FIXME: We need a more elegant way to register supported funcs */
static void ParCALSupportedFuncsInit(void )
{
	parcal_supported_funcs.push_back(NrmStringToQuark("dim_avg_n"));
	parcal_supported_funcs.push_back(NrmStringToQuark("dim_max_n"));
	parcal_supported_funcs.push_back(NrmStringToQuark("dim_min_n"));
	parcal_supported_funcs.push_back(NrmStringToQuark("abs"));
	parcal_supported_funcs.push_back(NrmStringToQuark("acos"));
	parcal_supported_funcs.push_back(NrmStringToQuark("max"));
	parcal_supported_funcs.push_back(NrmStringToQuark("min"));
	parcal_supported_funcs.push_back(NrmStringToQuark("dimsizes"));
	parcal_supported_funcs.push_back(NrmStringToQuark("asin"));
	parcal_supported_funcs.push_back(NrmStringToQuark("atan"));
	parcal_supported_funcs.push_back(NrmStringToQuark("atan2"));
	parcal_supported_funcs.push_back(NrmStringToQuark("sin"));
	parcal_supported_funcs.push_back(NrmStringToQuark("sinh"));
	parcal_supported_funcs.push_back(NrmStringToQuark("cos"));
	parcal_supported_funcs.push_back(NrmStringToQuark("exp"));
	parcal_supported_funcs.push_back(NrmStringToQuark("fabs"));
	parcal_supported_funcs.push_back(NrmStringToQuark("floor"));
	parcal_supported_funcs.push_back(NrmStringToQuark("ceil"));
	parcal_supported_funcs.push_back(NrmStringToQuark("log"));
	parcal_supported_funcs.push_back(NrmStringToQuark("log10"));
	parcal_supported_funcs.push_back(NrmStringToQuark("sqrt"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2vrG_Wrap"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2vrG"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2vrF_Wrap"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2vrF"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2vrA"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2dvG_Wrap"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2dvG"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2dvF_Wrap"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2dvF"));
	parcal_supported_funcs.push_back(NrmStringToQuark("uv2dvA"));
	parcal_supported_funcs.push_back(NrmStringToQuark("print"));
	parcal_supported_funcs.push_back(NrmStringToQuark("printVarSummary"));
	parcal_supported_funcs.push_back(NrmStringToQuark("getvardims"));
	parcal_supported_funcs.push_back(NrmStringToQuark("iscoord"));
	parcal_supported_funcs.push_back(NrmStringToQuark("getvaratts"));
	parcal_supported_funcs.push_back(NrmStringToQuark("isatt"));
	parcal_supported_funcs.push_back(NrmStringToQuark("typeof"));
	parcal_supported_funcs.push_back(NrmStringToQuark("asciiwrite"));
}

int ParCALIsSupportedFunc(char *fname)
{
	NclQuark fnameq = NrmStringToQuark(fname);
	for(std::vector<NclQuark>::iterator iter = parcal_supported_funcs.begin(); iter != parcal_supported_funcs.end(); ++iter){
		if(fnameq == *iter){
			return 1;
		}
	}
	return 0;
}

NhlErrorTypes ParCALInit(void )
{
	int ret;
	
	ret = MPI_Comm_dup(MPI_COMM_WORLD, &pcal_gcomm);
	if(ret != MPI_SUCCESS){
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error duping comm world");
		return NhlFATAL;
	}

	ret = MPI_Comm_rank(pcal_gcomm, &pcal_grank);
	if(ret != MPI_SUCCESS){
		MPI_Comm_free(&pcal_gcomm);
		NhlPError(NhlFATAL, NhlEUNKNOWN, "Error getting rank from dup comm");
		return NhlFATAL;
	}

	try{
		mb_gcore = new moab::Core();
	}catch(std::exception &e){
		NhlPError(NhlFATAL, ENOMEM, "Error creating MOAB instance");
		return NhlFATAL;
	}

	try{
		mb_gcore_pcomm = new moab::ParallelComm(mb_gcore, pcal_gcomm);
	}catch(std::exception &e){
		/* FIXME: Free comm, core ? */
		NhlPError(NhlFATAL, ENOMEM, "Error creating MOAB parallel communicator");
		return NhlFATAL;
	}

	time_gq = NrmStringToQuark("time");
	lev_gq = NrmStringToQuark("lev");

	ParCALToNCNclTypeMapInit();
	ParCALLocToMBTypeMapInit();
	ParCALSupportedFuncsInit();

	return NhlNOERROR;
}

#define MOAB_PART_METHOD "SQIJ"

/* FIXME: Use wr_status to define open mode */
void *ParCALOpen(NclQuark fname, int wr_status)
{
	ParCALHnd *pcal_hnd;
	std::vector<std::string> fnames;

	pcal_hnd = (ParCALHnd *) NclCalloc(1, sizeof(ParCALHnd));
    if(pcal_hnd == NULL){
        NhlPError(NhlFATAL, ENOMEM, NULL);
        return NULL;
    }

    fnames.push_back(std::string(NrmQuarkToString(fname)));

	parncl::MBMesh *mesh = create_mb_mesh(mb_gcore, fnames, pcal_gcomm, mb_gcore_pcomm, std::string(MOAB_PART_METHOD));

	pcal_hnd->mesh = mesh;
	mesh->inc_ref_cnt();
	pcal_hnd->fnames_idx = 0;

    return (void *)pcal_hnd;
} 

NhlErrorTypes ParCALMOpen(void **pcal_hnds, int npcal_hnds, NclMultiDValData files, int wr_status)
{
	std::vector<std::string> fnames;
	ParCALHnd *pcal_hnd;

	assert(npcal_hnds > 0);
	assert(npcal_hnds == files->multidval.totalelements);

	pcal_hnd = (ParCALHnd *) NclCalloc(1, sizeof(ParCALHnd));
	if(pcal_hnd == NULL){
        NhlPError(NhlFATAL, ENOMEM, NULL);
        return NhlFATAL;
	}

	for(int i=0; i<files->multidval.totalelements; i++){
		fnames.push_back(
			std::string(NrmQuarkToString( ((NclQuark *)files->multidval.val)[i] ))
		);
	}

	parncl::MBMesh *mesh = create_mb_mesh(mb_gcore, fnames, pcal_gcomm, mb_gcore_pcomm, std::string(MOAB_PART_METHOD));

	pcal_hnd->mesh = mesh;
	mesh->inc_ref_cnt();
	pcal_hnd->fnames_idx = 0;
	((void **)pcal_hnds)[0] = (void *)pcal_hnd;


	for(int i=1; i<npcal_hnds; i++){
		pcal_hnd = (ParCALHnd *) NclCalloc(1, sizeof(ParCALHnd));
		if(pcal_hnd == NULL){
			NhlPError(NhlFATAL, ENOMEM, NULL);
			return NhlFATAL;
		}
		/* Dup the pcal core handle */
		pcal_hnd->mesh = mesh;
		mesh->inc_ref_cnt();
		pcal_hnd->fnames_idx = i;
		((void **)pcal_hnds)[i] = (void *)pcal_hnd;
	}

    return NhlNOERROR;
}

int ParCALClose(void *pcal_hnd, NclQuark file)
{
	parncl::MBMesh *mesh = ((ParCALHnd *)pcal_hnd)->mesh;
	int retval = 0;

	assert(mesh != NULL);
    assert(mesh->get_ref_cnt() > 0);

	mesh->dec_ref_cnt();
    if(mesh->get_ref_cnt() == 0){
		delete mesh;
		/* We are done */
		retval = 0;
    }
    else{
		/*
		 * Currently there is no way to delete data corresponding to a file
		 * from the ParCAL instance.
		 */
		retval = 1;
    }
	NclFree(pcal_hnd);
	return retval;
} 

void ParCALFinalize(void )
{
	MPI_Barrier(pcal_gcomm);
	MPI_Comm_free(&pcal_gcomm);
	if(mb_gcore){
		//delete mb_gcore;
		mb_gcore = NULL;
	}
	return;
}

static nc_type ParCALMapToNC(const std::string &str)
{
	std::map<std::string, std::pair<nc_type, NclBasicDataTypes> >::iterator iter  = parcal_to_nc_gtype_map.find(str);
	if(iter != parcal_to_nc_gtype_map.end()){
		return (iter->second).first;
	}
	else{
		return NC_EBADTYPE;
	}
}

static NclBasicDataTypes ParCALMapToNcl(const std::string &str)
{
	std::map<std::string, std::pair<nc_type, NclBasicDataTypes> >::iterator iter  = parcal_to_nc_gtype_map.find(str);
	if(iter != parcal_to_nc_gtype_map.end()){
		return (iter->second).second;
	}
	else{
		return NCL_none;
	}
}

moab::EntityType ParCALLocToMB(const std::string &str)
{
	std::map<std::string, moab::EntityType>::iterator iter = parcal_loc_to_mb_gtype_map.find(str);
	if(iter != parcal_loc_to_mb_gtype_map.end()){
		return (iter->second);
	}
	else{
		return moab::MBMAXTYPE;
	}
}

template<typename T>
static void *ParCALCopyAttValFromVec(const pargal::pcatt& att, const std::string& att_name, int *lenp)
{
	T *att_val = NULL;
	std::vector<T> bvec;
	att.get_vals(att_name, bvec);

	*lenp = bvec.size();
	if(*lenp){
		att_val = (T *)NclMalloc(sizeof(T) * (*lenp));
		if(att_val){
			std::copy(bvec.begin(), bvec.end(), (T *)att_val);
		}
	}
	else{
		NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to copy att val of %s\n", att_name.c_str());
	}
	return (void *)att_val;
}

static void *ParCALGetAttVal(const pargal::pcatt& att, const std::string& att_name, NclBasicDataTypes type, int *lenp)
{
	void *att_val = NULL;
	std::string str;

	switch(type){
		case NCL_string:
			/* FIXME: What about strings with fillValues? */
			att.get_vals(att_name, str);
			att_val = NclMalloc(sizeof(NclQuark));
			if(att_val){
				*((NclQuark *) att_val) = NrmStringToQuark(str.c_str());
				*lenp = 1;
			}
			break;
		case NCL_short:
			att_val = ParCALCopyAttValFromVec<short>(att, att_name, lenp);
			break;
		case NCL_int:
			att_val = ParCALCopyAttValFromVec<int>(att, att_name, lenp);
			break;
		case NCL_float:
			att_val = ParCALCopyAttValFromVec<float>(att, att_name, lenp);
			break;
		case NCL_double:
			att_val = ParCALCopyAttValFromVec<double>(att, att_name, lenp);
			break;
		//case NCL_byte:
		default:
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to read att (%s), unknown type", att_name.c_str());
	}
	if(!att_val){
		NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to read att (%s)", att_name.c_str());
	}

	return att_val;
}

//static int ParCALGetAttInfo(ParCALCore *pcal_core_hnd, const pargal::pcatt&att, PNNCAttInqRecList **prec_list)
static int ParCALGetAttInfo(const pargal::pcatt&att, PNNCAttInqRecList **prec_list)
{
	int n_atts = 0, i;
	
	assert(prec_list != NULL);

	PNNCAttInqRecList *rec_list = *prec_list;

	std::set<std::string> att_names = att.get_att_names();

	PNNCAttInqRecList *cur_rec = rec_list;
	PNNCAttInqRecList *prev_rec = rec_list;
	i = 0;
	for(std::set<std::string>::iterator iter = att_names.begin(); iter != att_names.end(); ++iter, i++){
		if(cur_rec == NULL){
			cur_rec = (PNNCAttInqRecList *)NclCalloc(1, sizeof(PNNCAttInqRecList));
			if(cur_rec == NULL){
				NhlPError(NhlFATAL, ENOMEM, NULL);
				break;
			}
			if(prev_rec != NULL){
				prev_rec->next = cur_rec;
			}
			else{
				/* First node */
				rec_list = cur_rec;
			}
		}	
		/* Create and fill the Att Inq record */
		PNNCAttInqRec *att_inq = (PNNCAttInqRec *)NclCalloc(1, sizeof(PNNCAttInqRec));
		if(att_inq == NULL){
			if(cur_rec != rec_list){
				/* This is an incomplete record that was allocated here */
				NhlFree(cur_rec);
			}
			NhlPError(NhlFATAL, ENOMEM, NULL);
			break;
		}

		/* Fill-in att inq rec details */
		att_inq->att_num = i;
		att_inq->name = NrmStringToQuark((*iter).c_str());
		att_inq->varid = NC_GLOBAL;
		att_inq->data_type = ParCALMapToNcl(att.get_data_type(*iter));
		/* FIXME: Is int enough?  Do we need nc_size_t etc*/
		att_inq->value = ParCALGetAttVal(att, (*iter), att_inq->data_type, &(att_inq->len));

		cur_rec->att_inq = att_inq;
		n_atts++;
		prev_rec = cur_rec;
		cur_rec = cur_rec->next;
	}

	*prec_list = rec_list;
	return n_atts;
}

#define PARCAL_NC_MISSING_VAL_NAME "_FillValue"
static void *ParCALGetMissingVal(const pargal::fileinfo &pcal_finfo, const std::string &var_name)
{
	void *missing_val = NULL;
	const pargal::pcatt& att = pcal_finfo.get_var_att(var_name);
	std::set<std::string> att_names = att.get_att_names();
	std::set<std::string>::iterator iter = std::find(att_names.begin(), att_names.end(), std::string(PARCAL_NC_MISSING_VAL_NAME));

	if(iter != att_names.end()){
		NclBasicDataTypes ncl_type = ParCALMapToNcl(att.get_data_type(PARCAL_NC_MISSING_VAL_NAME));
		
		std::string missing_val_att_name = std::string(PARCAL_NC_MISSING_VAL_NAME);
		int len = 0;
		missing_val = ParCALGetAttVal(att, missing_val_att_name, ncl_type, &len); 
	}
	return missing_val;
}

/*
 * There is only one node in the rec list. Create nodes and fill in the
 * details
 */
int ParCALGetVarInfo(void *pcal_hnd, PNNCVarInqRecList **prec_list)
{
	parncl::MBMesh *mesh = ((ParCALHnd *) pcal_hnd)->mesh;
	int n_vars = 0, i = 0;

	assert(mesh != NULL);
	assert(prec_list != NULL);
	
	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();
	PNNCVarInqRecList *rec_list = *prec_list;
	
//	std::vector<std::string> var_names = pcal_core_hnd->pcore->get_var_names();
	std::set<std::string> var_names = pcal_finfo.var_names();

	PNNCVarInqRecList *cur_rec = rec_list;
	PNNCVarInqRecList *prev_rec = rec_list;
	for(std::set<std::string>::iterator iter = var_names.begin(); iter != var_names.end(); ++iter){
		if(cur_rec == NULL){
			cur_rec = (PNNCVarInqRecList *)NclCalloc(1, sizeof(PNNCVarInqRecList));
			if(cur_rec == NULL){
				NhlPError(NhlFATAL, ENOMEM, NULL);
				break;
			}
			if(prev_rec != NULL){
				prev_rec->next = cur_rec;
			}
			else{
				/* First node */
				rec_list = cur_rec;
			}
		}	
		/* Create and fill the Var Inq record */
		PNNCVarInqRec *var_inq = (PNNCVarInqRec *)NclCalloc(1, sizeof(PNNCVarInqRec));
		if(var_inq == NULL){
			if(cur_rec != rec_list){
				/* This is an incomplete record that was allocated here */
				NhlFree(cur_rec);
			}
			NhlPError(NhlFATAL, ENOMEM, NULL);
			break;
		}

		/* Variable name */
		var_inq->name = NrmStringToQuark((*iter).c_str());

		/* Variable dimensions */
		var_inq->n_dims = 0;
		try{
			std::vector<pargal::pcdim> var_dim_info = pcal_finfo.get_var_dims(*iter);

			i = 0;
			for(std::vector<pargal::pcdim>::iterator var_dim_info_iter = var_dim_info.begin(); (var_dim_info_iter != var_dim_info.end()) && (i < MAX_VAR_DIMS); ++var_dim_info_iter, i++){
				/*
				 * FIXME: ParCAL currently does not provide info on dim names
				 * - So ParCAL interface needs fixing here. Till then we just 
				 * copy the dim value instead of the whole dim info or dimension id
				 */
				var_inq->dim_info[i].name = NrmStringToQuark((*var_dim_info_iter).name().c_str());
				var_inq->dim_info[i].dim_id = mesh_finfo.get_dim_id((*var_dim_info_iter).name());;
				var_inq->dim_info[i].size = (*var_dim_info_iter).size();	
				var_inq->n_dims++;
			}

			/* Variable datatype - default to NC_DOUBLE */
			/* FIXME: This should be provided by ParCAL */
			var_inq->data_type = NC_DOUBLE;
		} catch (pargal::pargal_except &e){
			std::cerr << e.what() << std::endl;
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to get dims for %s", (*iter).c_str());
		}

		/* Variable attributes */
		var_inq->n_atts = 0;
		var_inq->att_list = NULL;

		try{
			const pargal::pcatt& var_att = pcal_finfo.get_var_att(*iter);
			var_inq->n_atts = ParCALGetAttInfo(var_att, &(var_inq->att_list));
		} catch (pargal::pargal_except &e){
			std::cerr << e.what() << std::endl;
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to get atts for %s", (*iter).c_str());
		}

		cur_rec->var_inq = var_inq;
		n_vars++;
		prev_rec = cur_rec;
		cur_rec = cur_rec->next;
	}

	*prec_list = rec_list;
	return n_vars;	
}

int ParCALGetDimInfo(void *pcal_hnd, PNNCDimInqRecList **prec_list)
{
	parncl::MBMesh *mesh = ((ParCALHnd *) pcal_hnd)->mesh;
	int n_dims = 0;
	
	assert(mesh != NULL);
	assert(prec_list != NULL);

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	PNNCDimInqRecList *rec_list = *prec_list;

	std::set<std::string> dim_names = pcal_finfo.dim_names();

	PNNCDimInqRecList *cur_rec = rec_list;
	PNNCDimInqRecList *prev_rec = rec_list;
	for(std::set<std::string>::iterator iter = dim_names.begin(); iter != dim_names.end(); ++iter){
		if(cur_rec == NULL){
			cur_rec = (PNNCDimInqRecList *)NclCalloc(1, sizeof(PNNCDimInqRecList));
			if(cur_rec == NULL){
				NhlPError(NhlFATAL, ENOMEM, NULL);
				break;
			}
			if(prev_rec != NULL){
				prev_rec->next = cur_rec;
			}
			else{
				/* First node */
				rec_list = cur_rec;
			}
		}	
		/* Create and fill the Dim Inq record */
		PNNCDimInqRec *dim_inq = (PNNCDimInqRec *)NclCalloc(1, sizeof(PNNCDimInqRec));
		if(dim_inq == NULL){
			if(cur_rec != rec_list){
				/* This is an incomplete record that was allocated here */
				NhlFree(cur_rec);
			}
			NhlPError(NhlFATAL, ENOMEM, NULL);
			break;
		}

		/* Fill-in dim inq rec details */
		//dim_inq->dimid = ParCALCreateDimId(pcal_core_hnd, *iter);
		dim_inq->dimid = mesh_finfo.insert_dim_name(*iter);

		dim_inq->name = NrmStringToQuark((*iter).c_str());

		try{
			std::vector<pargal::pcdim> dim_info = pcal_finfo.get_var_dims(*iter);
			if(dim_inq->name != time_gq){
				dim_inq->size = (int )dim_info[0].size();
			}
			else{
				/* The "time" dim is uniformly distributed across files */
				dim_inq->size = dim_info[0].size()/mesh_finfo.get_num_nc_files();
			}
			dim_inq->is_unlimited = (dim_info[0].is_unlimited()) ? 1 : 0;
		}catch(pargal::pargal_except &e){
			std::cerr << e.what() << std::endl;
			NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to get dim info for %s", (*iter).c_str());
		}

		cur_rec->dim_inq = dim_inq;
		n_dims++;
		prev_rec = cur_rec;
		cur_rec = cur_rec->next;
	}

	*prec_list = rec_list;
	return n_dims;
}


int ParCALGetFileAttInfo(void *pcal_hnd, PNNCAttInqRecList **prec_list)
{
	parncl::MBMesh *mesh = ((ParCALHnd *) pcal_hnd)->mesh;
	
	assert(mesh != NULL);
	assert(prec_list != NULL);

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	try{
		const pargal::pcatt& att = pcal_finfo.get_att();
		return ParCALGetAttInfo(att, prec_list);
	}catch(pargal::pargal_except &e){
		std::cerr << e.what() << std::endl;
		NhlPError(NhlWARNING, NhlEUNKNOWN, "Unable to get att info from file");
		return 0;
	}

}

/* FIXME: These funcs - need to be moved to other files */
void *ParCALUVToVort(void *u_pvar_shnd, void *v_pvar_shnd)
{
	return parncl::ParCALTwoMDIOp(pargal::vorticity, u_pvar_shnd, v_pvar_shnd);
}

void *ParCALUVToDiv(void *u_pvar_shnd, void *v_pvar_shnd)
{
	return parncl::ParCALTwoMDIOp(pargal::divergence, u_pvar_shnd, v_pvar_shnd);
}

/* FIXME: We should be using ng_size_t where appropriate, - dim_sizes ? */
/* This function creates a new variable with the same partition and 
 * attributes as the input variable
 */
// FIXME: How can we get this functionality to the meshval class?
void *ParCALCreateVarFromVar(void *from_pvar_shnd, NclSelectionRecord *sel)
{
	parncl::ParCALVarSHnd *from_pv_shnd = (parncl::ParCALVarSHnd *)from_pvar_shnd;
	parncl::ParCALVarSHnd *pv_shnd = NULL;

	if(sel == NULL){
		parncl::MBMesh *mesh = from_pv_shnd->get_mesh();
		moab::Interface *mb_core = mesh->get_mb_core();

		size_t var_nlevs = from_pv_shnd->get_num_val_levs();
		const std::vector<int>& var_tsteps = from_pv_shnd->get_pcvar()->get_tsteps();
		const std::vector<pargal::pcdim>& var_dims = from_pv_shnd->get_pcvar()->get_dims();

		moab::EntityType ent_type = ParCALLocToMB(from_pv_shnd->get_pcvar()->get_loc());
		size_t sz = PNclMBGetChunkSz(mesh, ent_type);

		NclSelectionRecord *sel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
		if(!sel){
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Unable to allocate mem for var sel");
			return NULL;
		}

		from_pv_shnd->get_sel(sel);
		if(sel->n_entries == 0){
			NclFree(sel);
			sel = NULL;
		}

		parncl::ParCALMeshValSPtr sptr = parncl::ParCALMeshVal::create_mesh_val(mesh, mb_core, sz, var_tsteps, var_dims, var_nlevs, from_pv_shnd->get_data_type(), from_pv_shnd->get_pcvar()->get_loc(), from_pv_shnd->get_missing_val(), sel);
		pv_shnd = new parncl::ParCALVarSHnd(sptr);
	}
	else{
		pv_shnd = new parncl::ParCALVarSHnd(*from_pv_shnd, sel);
	}

	return (void *)pv_shnd;
}

void *ParCALDimAvg(NclMultiDValData md_val, int *dims, ng_size_t n_dims, NclSelectionRecord *sel)
{
	/* FIXME: Support sel */
	void *p_var_shnd = md_val->data.shnd;
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	parncl::MBMesh *mesh = pv_shnd->get_mesh();
	moab::Interface *mb_core = mesh->get_mb_core();

	/* FIXME: Do we intend to support floats ? */
	if(n_dims != 0){
		std::string dim_avg_name;
		pargal::pcvar *cur_pcvar = pv_shnd->get_pcvar();
		const std::string &cur_pcvar_loc = pv_shnd->get_pcvar()->get_loc();
		const std::vector<pargal::pcdim>& var_dims = cur_pcvar->get_dims();
		const std::vector<int>& var_tsteps = cur_pcvar->get_tsteps();
		std::vector<int> var_one_tstep(1, 0);
		moab::DataType dt = PNclNCMapToMB(pv_shnd->get_data_type());

		/* FIXME: Use the new pargal interface for multiple dims */
		for(size_t i=0; i<n_dims; i++){
			dim_avg_name = mesh->get_ustr();
			size_t ind = dims[i];
			const std::vector<size_t> inds(1, ind);
			double pv_shnd_missing_val = 0.0;
			pv_shnd->get_missing_val(pv_shnd_missing_val);

			pargal::pcvar *dim_avg_var = new pargal::pcvar(dim_avg_name, dt,
										var_dims,
										pv_shnd->get_pcal_finfo(),
										(ind == 0) ? var_one_tstep : var_tsteps,
										true, cur_pcvar_loc, false, pv_shnd_missing_val);

			pargal::dim_avg_n(*(cur_pcvar), *(dim_avg_var), inds);
			cur_pcvar = dim_avg_var;
			/* FIXME: Delete temp pv_shnds */
		}

		/* Add selector to note "real/valid" dimensions of var */
		NclSelectionRecord *pcsel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
		if(pcsel){
			size_t n_md_dims = md_val->multidval.n_dims;
			pcsel->n_entries = n_md_dims;
			/* Initialize the sel entries */
			for(size_t i=0; i<n_md_dims; i++){
				/* FIXME: Slightly different take on SUB_ALL where start & finish
				 * is explicitly specified
				 */
				pcsel->selection[i].sel_type = Ncl_SUB_ALL;
				pcsel->selection[i].u.sub.start = 0;
				pcsel->selection[i].u.sub.finish = md_val->multidval.dim_sizes[i] - 1;
				pcsel->selection[i].u.sub.stride = 1;
			}
			for(size_t i=0; i<n_dims; i++){
				size_t ind = dims[i];
				pcsel->selection[ind].sel_type = Ncl_SUBSCR;
				pcsel->selection[ind].u.sub.start = 0;
				pcsel->selection[ind].u.sub.finish = 0;
				pcsel->selection[ind].u.sub.stride = 1;
			}
		}

		parncl::ParCALMeshValSPtr sptr = parncl::ParCALMeshVal::create_mesh_val(cur_pcvar, NC_DOUBLE, false, pv_shnd->get_missing_val());
		moab::EntityType ent_type = ParCALLocToMB(pv_shnd->get_pcvar()->get_loc());
		size_t mb_chunk_sz = PNclMBGetChunkSz(mesh, ent_type);
		sptr->set_mesh_info(mesh, mb_core, mb_chunk_sz, parncl::ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
		parncl::ParCALVarSHnd *pv_shnd_dim_avg = new parncl::ParCALVarSHnd(sptr, pcsel);

		return (void *)pv_shnd_dim_avg;
	}
	return NULL;
}

typedef void (*ParCALMinMaxTemplDimNFunc) (pargal::pcvar &, pargal::pcvar &, const std::vector<size_t> &, bool);
typedef double (*ParCALMinMaxTemplDimZFunc) (pargal::pcvar &, bool);

/* FIXME: Passing in a valid min_var shnd does not work yet */
void *ParCALMinMaxTempl(ParCALMinMaxTemplDimZFunc dim_z_func, ParCALMinMaxTemplDimNFunc dim_n_func, NclMultiDValData md_val, int *dims, ng_size_t n_dims, NclSelectionRecord *sel, void *res_vals, ng_size_t res_vals_len, NclBasicDataTypes *otype)
{
	/* FIXME: Support sel */
	void *p_var_shnd = md_val->data.shnd;
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	parncl::MBMesh *mesh = pv_shnd->get_mesh();
	moab::Interface *mb_core = mesh->get_mb_core();

	/* FIXME: Do we intend to support floats ? */
	if(n_dims == 0){
		double val = (*dim_z_func)(*(pv_shnd->get_pcvar()), false);
		memcpy(res_vals, (void *)&val, res_vals_len);
		*otype = NCL_double;
		return res_vals;
	}
	else{
		std::string res_var_name;
		pargal::pcvar *cur_pcvar = pv_shnd->get_pcvar();
		const std::string &cur_pcvar_loc = pv_shnd->get_pcvar()->get_loc();
		const std::vector<pargal::pcdim>& var_dims = cur_pcvar->get_dims();
		const std::vector<int>& var_tsteps = cur_pcvar->get_tsteps();
		std::vector<int> var_one_tstep(1, 0);
		moab::DataType dt = PNclNCMapToMB(pv_shnd->get_data_type());

		/* FIXME: Use the new pargal interface for multiple dims */
		for(size_t i=0; i<n_dims; i++){
			res_var_name = mesh->get_ustr();
			size_t ind = dims[i];
			std::vector<size_t> inds(1, ind);
			double pv_shnd_missing_val = 0.0;
			pv_shnd->get_missing_val(pv_shnd_missing_val);

			pargal::pcvar *res_var = new pargal::pcvar(res_var_name, dt,
										var_dims,
										pv_shnd->get_pcal_finfo(),
										(ind == 0) ? var_one_tstep : var_tsteps,
										true, cur_pcvar_loc, false, pv_shnd_missing_val);

			(*dim_n_func)(*(cur_pcvar), *(res_var), inds, false);
			cur_pcvar = res_var;
			/* FIXME: Delete temp pv_shnds */
		}

		/* Add selector to note "real/valid" dimensions of var */
		NclSelectionRecord *pcsel = (NclSelectionRecord *)NclCalloc(1, sizeof(NclSelectionRecord));
		if(pcsel){
			size_t n_md_dims = md_val->multidval.n_dims;
			pcsel->n_entries = n_md_dims;
			/* Initialize the sel entries */
			for(size_t i=0; i<n_md_dims; i++){
				/* FIXME: Slightly different take on SUB_ALL where start & finish
				 * is explicitly specified
				 */
				pcsel->selection[i].sel_type = Ncl_SUB_ALL;
				pcsel->selection[i].u.sub.start = 0;
				pcsel->selection[i].u.sub.finish = md_val->multidval.dim_sizes[i] - 1;
				pcsel->selection[i].u.sub.stride = 1;
			}
			for(size_t i=0; i<n_dims; i++){
				size_t ind = dims[i];
				pcsel->selection[ind].sel_type = Ncl_SUBSCR;
				pcsel->selection[ind].u.sub.start = 0;
				pcsel->selection[ind].u.sub.finish = 0;
				pcsel->selection[ind].u.sub.stride = 1;
			}
		}

		parncl::ParCALMeshValSPtr sptr = parncl::ParCALMeshVal::create_mesh_val(cur_pcvar, NC_DOUBLE, false, pv_shnd->get_missing_val());
		moab::EntityType ent_type = ParCALLocToMB(pv_shnd->get_pcvar()->get_loc());
		size_t mb_chunk_sz = PNclMBGetChunkSz(mesh, ent_type);
		sptr->set_mesh_info(mesh, mb_core, mb_chunk_sz, parncl::ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
		parncl::ParCALVarSHnd *pv_shnd_res = new parncl::ParCALVarSHnd(sptr, pcsel);

		*otype = NCL_pdouble;
		return (void *)pv_shnd_res;
	}
}

void *ParCALMin(NclMultiDValData md_val, int *dims, ng_size_t n_dims, NclSelectionRecord *sel, void *min_vals, ng_size_t min_vals_len, NclBasicDataTypes *otype)
{
	return ParCALMinMaxTempl(pargal::min, pargal::dim_min_n, md_val, dims, n_dims, sel, min_vals, min_vals_len, otype);
}

/* FIXME: Passing in a valid max_var shnd does not work yet */
void *ParCALMax(NclMultiDValData md_val, int *dims, ng_size_t n_dims, NclSelectionRecord *sel, void *max_vals, ng_size_t max_vals_len, NclBasicDataTypes *otype)
{
	return ParCALMinMaxTempl(pargal::max, pargal::dim_max_n, md_val, dims, n_dims, sel, max_vals, max_vals_len, otype);
}

void *ParCALDupVar(void *p_var_shnd)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	
	parncl::ParCALVarSHnd *dup_pv_shnd = new parncl::ParCALVarSHnd(*pv_shnd);
	/* Copy by ref */

	return (void *)dup_pv_shnd;
}

void *ParCALCacheVar(void **pcal_hnds, int npcal_hnds, NclQuark var_nameq, NclFVarSelRec *sel, int nsel)
{
	ParCALHnd *pcal_hnd;
	parncl::MBMesh *mesh = NULL;

	assert((npcal_hnds > 0) && (nsel >= 0));

	pcal_hnd = (ParCALHnd *) ((void **)pcal_hnds)[0];
	assert(pcal_hnd != NULL);
	mesh = pcal_hnd->mesh;
	moab::Interface *mb_core = mesh->get_mb_core();

	parncl::MeshFInfo &mesh_finfo = mesh->get_mesh_finfo();
	pargal::fileinfo &pcal_finfo = mesh_finfo.get_pcal_finfo();

	/* Create a parcal var */
	std::string var_name(NrmQuarkToString(var_nameq));

	const std::vector<pargal::pcdim>& var_dims = pcal_finfo.get_var_dims(var_name);
    /* We assume that time dim is uniformly distributed across files */
    int ntsteps_in_file = 1;
    std::vector<int> var_tsteps;

    if(NrmStringToQuark(var_dims[0].name().c_str()) == time_gq){
        ntsteps_in_file = var_dims[0].size()/mesh_finfo.get_num_nc_files();
        for(int i=0; i<npcal_hnds; i++){
            pcal_hnd = (ParCALHnd *) pcal_hnds[i];
            assert(pcal_hnd != NULL);
            for(int j=0; j<ntsteps_in_file; j++){
                var_tsteps.push_back(ntsteps_in_file * pcal_hnd->fnames_idx + j);
            }
        }
    }
    else{
        var_tsteps.push_back(0);
    }

	/* FIXME: If we have a var selector use Scd interface to get the var */

	/* Don't assume double */
	moab::DataType dt = PNclNCMapToMB(NC_DOUBLE);
	pargal::pcvar *pcvar = new pargal::pcvar(var_name, dt, var_dims, mesh_finfo.get_pcal_finfo(), var_tsteps, false);

	void *missing_val = ParCALGetMissingVal(mesh_finfo.get_pcal_finfo(), var_name);
	moab::EntityType ent_type = ParCALLocToMB(pcvar->get_loc());
	size_t mb_chunk_sz = PNclMBGetChunkSz(mesh, ent_type);
	parncl::ParCALMeshValSPtr sptr = parncl::ParCALMeshVal::create_mesh_val(pcvar, NC_DOUBLE, true, missing_val);
	sptr->set_mesh_info(mesh, mb_core, mb_chunk_sz, parncl::ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
	parncl::ParCALVarSHnd *pv_shnd = new parncl::ParCALVarSHnd(sptr);	

	return (void *)pv_shnd;
}

NhlErrorTypes ParCALFreeVar(void *p_var_shnd)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	delete pv_shnd;

	return NhlNOERROR;
}

NclTypeClass ParCALNCToNclClass(nc_type type, ParCALValType val_type)
{
	NclTypeClass otype_class = NULL;
	switch(type){
		case NC_DOUBLE:
			if(val_type == PARCAL_VAL_LOCAL){
				otype_class = (NclTypeClass) nclTypepdoubleClass;
			}
			else{
				if(pcal_grank == 0){
					/* Only rank 0 has the gathered data */
					otype_class = (NclTypeClass) nclTypedoubleClass;
				}
				else{
					/* FIXME: For now treat this val as double and declare
					 * val as unusable
					 */
					otype_class = (NclTypeClass) nclTypedoubleClass;
				}
			}
			break;
		default:
			NhlPError(NhlFATAL, NhlEUNKNOWN, "Error mapping NC type to Ncl class type");
	}
	return otype_class;
}

int ParCALGetNumChunks(void *p_var_shnd)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	return(pv_shnd->get_num_val_chunks());
}

/* FIXME: The returned ptr is only valid till the next call to ParCALGetVal() . We
 * should use an iterator instead and the caller should explicitly invalidate this ptr.
 */
void *ParCALGetVal(void *p_var_shnd, NclSelectionRecord *sel, ng_size_t req_chunk, ParCALValType val_type, NclTypeClass *out_val_type, ParCALAccessMode mode, ng_size_t *sz)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	void *p_data = NULL;
	size_t p_data_sz = 0;

	/* FIXME: Add support for selecting an arbitrary data slice */
	assert(sel == NULL);
	if(val_type == PARCAL_VAL_GLOBAL){
		req_chunk = parncl::PARCALVAR_NCHUNK_ALL;
	}

	p_data = pv_shnd->get_val(sel, req_chunk, val_type, mode, p_data_sz);
	if(val_type == PARCAL_VAL_LOCAL){
		if(out_val_type){
			*out_val_type = ParCALNCToNclClass(pv_shnd->get_data_type(), PARCAL_VAL_LOCAL);
		}
	}
	else{
		if(out_val_type){
			*out_val_type = ParCALNCToNclClass(pv_shnd->get_data_type(), PARCAL_VAL_GLOBAL);
		}
	}
	*sz = p_data_sz;

	return p_data;
}

void ParCALClearValCache(void *p_var_shnd, ng_size_t req_chunk, ParCALValType val_type)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	if(val_type == PARCAL_VAL_GLOBAL){
		req_chunk = parncl::PARCALVAR_NCHUNK_ALL;
	}
	pv_shnd->delete_val(req_chunk);	
}

void ParCALVarSetDecloneLevel(void *p_var_shnd, int caller_level)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	pv_shnd->set_declone_caller_level(caller_level);
}

int ParCALVarGetDecloneLevel(void *p_var_shnd)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	return pv_shnd->get_declone_caller_level();
}

void ParCALVarDeclone(void *p_var_shnd, NclTypeClass *declone_val_type)
{
	parncl::ParCALVarSHnd *pv_shnd = (parncl::ParCALVarSHnd *)p_var_shnd;
	if(declone_val_type){
		*declone_val_type = ParCALNCToNclClass(pv_shnd->get_data_type(), PARCAL_VAL_LOCAL);
	}
} 
