/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 * Development of this software was sponsored by the Office of Biological  *
 * and Environmental Research of the U.S. Department of Energy's Office    *
 * of Science.                                                             *
 *                                                                         *
 * Copyright 2011 UChicago Argonne, LLC.                                   *
 *                                                                         *
 * Licensed under the Apache License, Version 2.0 (the "License"); you     *
 * may not use this file except in compliance with the License.  You may   *
 * obtain a copy of the License at                                         * 
 *                                                                         *
 *   http://www.apache.org/licenses/LICENSE-2.0                            *
 *                                                                         *
 * Unless required by applicable law or agreed to in writing, software     *
 * distributed under the License is distributed on an "AS IS" BASIS,       *
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or         *
 * implied.  See the License for the specific language governing           *
 * permissions and limitations under the License.                          *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
#include <iostream>
#include <string>
#include <vector>
#include <map>

#define MPICH_SKIP_MPICXX
#define OMPI_SKIP_MPICXX
extern "C" {
#include "ParCALFuncs.h"
}
#include "mpi.h"
#ifndef MPI_INCLUDED
#define MPI_INCLUDED
#endif
/*
#include "parcal/fileinfo.hpp"
#include "parcal/pcdim.hpp"
#include "parcal/algorithm.hpp"
#include "parcal/exception.hpp"
*/
#include "moab/Core.hpp"
#include "pargal/pargal.hpp"
#include "ParCALCore.hpp"

#include "ParCALVarSHnd.hpp"
#include "MBMesh.hpp"
#include "ParCALMeshVal.hpp"
#include "PNclMBUtils.hpp"
#include "ParCALIAlgo.hpp"

namespace parncl{

void *ParCALTwoMDIOp(pargal_two_inarg_func_t func, void *arg1_pvar_shnd, void *arg2_pvar_shnd)
{
    parncl::ParCALVarSHnd *arg1 = (parncl::ParCALVarSHnd *)arg1_pvar_shnd;
    parncl::ParCALVarSHnd *arg2 = (parncl::ParCALVarSHnd *)arg2_pvar_shnd;
    parncl::ParCALVarSHnd *res = (parncl::ParCALVarSHnd *)NULL;

    parncl::MBMesh *mesh = arg1->get_mesh();
	moab::Interface *mb_core = mesh->get_mb_core();

    std::string res_var_name = mesh->get_ustr();

    const std::vector<pargal::pcdim>& arg1_dims = arg1->get_pcvar()->get_dims();

    const std::vector<int>& tsteps = arg1->get_pcvar()->get_tsteps();

    const std::string &res_var_loc = arg1->get_pcvar()->get_loc();

    /* FIXME: Shouldn't there be a special var to say all timesteps */
    moab::DataType dt = PNclNCMapToMB(arg1->get_data_type());
    double arg1_missing_val = 0.0;
    arg1->get_missing_val(arg1_missing_val);
    pargal::pcvar *res_var = new pargal::pcvar(res_var_name, dt,
                                arg1_dims, arg1->get_pcal_finfo(), tsteps, true, res_var_loc,
                                false, arg1_missing_val);

    (*func)(*arg1->get_pcvar(), *arg2->get_pcvar(), *res_var);

    parncl::ParCALMeshValSPtr sptr = parncl::ParCALMeshVal::create_mesh_val(res_var, arg1->get_data_type(), false, arg1->get_missing_val());
    moab::EntityType ent_type = ParCALLocToMB(res_var_loc);
    size_t sz = PNclMBGetChunkSz(mesh, ent_type);
    sptr->set_mesh_info(mesh, mb_core, sz, parncl::ParCALMeshVal::MESHVAL_LOC_UNKNOWN);
    res = new parncl::ParCALVarSHnd(sptr);

    return (void *)res;
}

} // namespace parncl
